from os.path import relpath

from coala_utils.decorators import enforce_signature, get_public_members
from coalib.results.SourcePosition import SourcePosition
from coalib.results.TextRange import TextRange
from coalib.results.AbsolutePosition import AbsolutePosition


class SourceRange(TextRange):

    @enforce_signature
    def __init__(self,
                 start: SourcePosition,
                 end: (SourcePosition, None)=None):
        """
        Creates a new SourceRange.

        :param start:       A SourcePosition indicating the start of the range.
        :param end:         A SourcePosition indicating the end of the range.
                            If ``None`` is given, the start object will be used
                            here. end must be in the same file and be greater
                            than start as negative ranges are not allowed.
        :raises TypeError:  Raised when
                            - start is not of type SourcePosition.
                            - end is neither of type SourcePosition, nor is it
                              None.
        :raises ValueError: Raised when file of start and end mismatch.
        """
        TextRange.__init__(self, start, end)

        if self.start.file != self.end.file:
            raise ValueError('File of start and end position do not match.')

    @classmethod
    def from_values(cls,
                    file,
                    start_line=None,
                    start_column=None,
                    end_line=None,
                    end_column=None):
        start = SourcePosition(file, start_line, start_column)
        if end_line or (end_column and end_column > start_column):
            end = SourcePosition(file, end_line if end_line else start_line,
                                 end_column)
        else:
            end = None

        return cls(start, end)

    @classmethod
    def from_clang_range(cls, range):
        """
        Creates a SourceRange from a clang SourceRange object.

        :param range: A cindex.SourceRange object.
        """
        return cls.from_values(range.start.file.name,
                               range.start.line,
                               range.start.column,
                               range.end.line,
                               range.end.column)

    @classmethod
    @enforce_signature
    def from_absolute_position(cls,
                               file: str,
                               position_start: AbsolutePosition,
                               position_end: (AbsolutePosition, None)=None):
        """
        Creates a SourceRange from a start and end positions.

        :param file:           Name of the file.
        :param position_start: Start of range given by AbsolutePosition.
        :param position_end:   End of range given by AbsolutePosition or None.
        """
        start = SourcePosition(file, position_start.line, position_start.column)
        end = None
        if position_end:
            end = SourcePosition(file, position_end.line, position_end.column)
        return cls(start, end)

    @property
    def file(self):
        return self.start.file

    @enforce_signature
    def renamed_file(self, file_diff_dict: dict):
        """
        Retrieves the filename this source range refers to while taking the
        possible file renamings in the given file_diff_dict into account:

        :param file_diff_dict: A dictionary with filenames as key and their
                               associated Diff objects as values.
        """
        diff = file_diff_dict.get(self.file)
        if diff is None:
            return self.file

        return diff.rename if diff.rename is not False else self.file

    def expand(self, file_contents):
        """
        Passes a new SourceRange that covers the same area of a file as this
        one would. All values of None get replaced with absolute values.

        values of None will be interpreted as follows:
        self.start.line is None:   -> 1
        self.start.column is None: -> 1
        self.end.line is None:     -> last line of file
        self.end.column is None:   -> last column of self.end.line

        :param file_contents: File contents of the applicable file
        :return:              TextRange with absolute values
        """
        tr = TextRange.expand(self, file_contents)

        return SourceRange.from_values(self.file,
                                       tr.start.line,
                                       tr.start.column,
                                       tr.end.line,
                                       tr.end.column)

    def __json__(self, use_relpath=False):
        _dict = get_public_members(self)
        if use_relpath:
            _dict['file'] = relpath(_dict['file'])
        return _dict

    def __str__(self):
        """
        Creates a string representation of the SourceRange object.

        If the whole file is affected, then just the filename is shown.

        >>> str(SourceRange.from_values('test_file', None, None, None, None))
        '...test_file'

        If the whole line is affected, then just the filename with starting
        line number and ending line number is shown.

        >>> str(SourceRange.from_values('test_file', 1, None, 2, None))
        '...test_file: L1 : L2'

        This is the general case where particular column and line are
        specified. It shows the starting line and column and ending line
        and column, with filename in the beginning.

        >>> str(SourceRange.from_values('test_file', 1, 1, 2, 1))
        '...test_file: L1 C1 : L2 C1'
        """
        if self.start.line is None and self.end.line is None:
            format_str = '{0.start.file}'
        elif self.start.column is None and self.end.column is None:
            format_str = '{0.start.file}: L{0.start.line} : L{0.end.line}'
        else:
            format_str = ('{0.start.file}: L{0.start.line} C{0.start.column}' +
                          ' : L{0.end.line} C{0.end.column}')

        return format_str.format(self)

    def __contains__(self, item):
        return (super().__contains__(item) and
                self.start.file == item.start.file)



from coalib.results.Result import Result
from coalib.results.result_actions.ResultAction import ResultAction

from coala_utils.decorators import enforce_signature


class PrintMoreInfoAction(ResultAction):

    @staticmethod
    @enforce_signature
    def is_applicable(result: Result, original_file_dict, file_diff_dict):
        if result.additional_info != '':
            return True
        return 'There is no additional info.'

    def apply(self, result, original_file_dict, file_diff_dict):
        """
        Print more info
        """
        print(result.additional_info)

        return file_diff_dict



import logging

from coala_utils.string_processing import escape

from coalib.bearlib.abstractions.SectionCreatable import SectionCreatable
from coalib.bearlib.languages import Language
from coalib.settings.Setting import Setting


class LanguageDefinition(SectionCreatable):
    """
    **This class is deprecated!** Use the `Language` class instead.

    A Language Definition holds constants which may help parsing the language.
    If you want to write a bear you'll probably want to use those definitions
    to keep your bear independent of the semantics of each language.

    You can easily get your language definition by just creating it with
    the name of the language desired:

    >>> list(LanguageDefinition("cpp")['extensions'])
    ['.c', '.cpp', '.h', '.hpp']

    For some languages aliases exist, the name is case insensitive; they will
    behave just like before and return settings:

    >>> dict(LanguageDefinition('C++')['comment_delimiter'])
    {'//': ''}
    >>> dict(LanguageDefinition('C++')['string_delimiters'])
    {'"': '"'}

    If no language exists, you will get a ``FileNotFoundError``:

    >>> LanguageDefinition("BULLSHIT!")
    Traceback (most recent call last):
     ...
    FileNotFoundError

    Custom coalangs are no longer supported. You can simply register your
    languages to the Languages decorator. When giving a custom coalang
    directory a warning will be emitted and it will attempt to load the given
    Language anyway through conventional means:

    >>> LanguageDefinition("custom", coalang_dir='somewhere')
    Traceback (most recent call last):
     ...
    FileNotFoundError

    If you need a custom language, just go like this:

    >>> @Language
    ... class MyLittlePony:
    ...     color = 'green'
    ...     legs = 5
    >>> int(LanguageDefinition('mylittlepony')['legs'])
    5

    But seriously, just use `Language` - and mind that it's already typed:

    >>> Language['mylittlepony'].get_default_version().legs
    5
    """

    def __init__(self, language: str, coalang_dir=None):
        """
        Creates a new LanguageDefinition object from file.

        :param language:           The actual language (e.g. C++).
        :param coalang_dir:        Path to directory with coalang language
                                   definition files. This replaces the default
                                   path if given.
        :raises FileNotFoundError: Raised when no definition is available for
                                   the given language.
        """
        logging.debug('LanguageDefinition has been deprecated! '
                      'Use `coalib.bearlib.languages.Language` instead.')
        if coalang_dir:
            logging.error(
                'LanguageDefinition has been deprecated. The `coalang_dir` '
                'functionality is not available anymore.')
        try:
            self.lang = Language[language].get_default_version()
        except AttributeError:
            raise FileNotFoundError

    def __getitem__(self, item):
        value = getattr(self.lang, item)
        if isinstance(value, (list, tuple)):
            value = Setting(item, ', '.join(escape(val, ',') for val in value))
        elif isinstance(value, dict):
            value = Setting(item, ', '.join(
                escape(key, ':,') + ': ' + escape(val, ':,')
                for key, val in value.items()))
        else:
            value = Setting(item, str(value))
        return value

    def __contains__(self, item):
        return item in self.lang.attributes



from contextlib import contextmanager
import shlex
from subprocess import PIPE, Popen


@contextmanager
def run_interactive_shell_command(command, **kwargs):
    """
    Runs a single command in shell and provides stdout, stderr and stdin
    streams.

    This function creates a context manager that sets up the process (using
    ``subprocess.Popen()``), returns to caller and waits for process to exit on
    leaving.

    By default the process is opened in ``universal_newlines`` mode and creates
    pipes for all streams (stdout, stderr and stdin) using ``subprocess.PIPE``
    special value. These pipes are closed automatically, so if you want to get
    the contents of the streams you should retrieve them before the context
    manager exits.

    >>> with run_interactive_shell_command(["echo", "TEXT"]) as p:
    ...     stdout = p.stdout
    ...     stdout_text = stdout.read()
    >>> stdout_text
    'TEXT\\n'
    >>> stdout.closed
    True

    Custom streams provided are not closed except of ``subprocess.PIPE``.

    >>> from tempfile import TemporaryFile
    >>> stream = TemporaryFile()
    >>> with run_interactive_shell_command(["echo", "TEXT"],
    ...                                    stdout=stream) as p:
    ...     stderr = p.stderr
    >>> stderr.closed
    True
    >>> stream.closed
    False

    :param command: The command to run on shell. This parameter can either
                    be a sequence of arguments that are directly passed to
                    the process or a string. A string gets splitted beforehand
                    using ``shlex.split()``. If providing ``shell=True`` as a
                    keyword-argument, no ``shlex.split()`` is performed and the
                    command string goes directly to ``subprocess.Popen()``.
    :param kwargs:  Additional keyword arguments to pass to
                    ``subprocess.Popen`` that are used to spawn the process.
    :return:        A context manager yielding the process started from the
                    command.
    """
    if not kwargs.get('shell', False) and isinstance(command, str):
        command = shlex.split(command)

    args = {'stdout': PIPE,
            'stderr': PIPE,
            'stdin': PIPE,
            'universal_newlines': True}
    args.update(kwargs)

    process = Popen(command, **args)
    try:
        yield process
    finally:
        if args['stdout'] is PIPE:
            process.stdout.close()
        if args['stderr'] is PIPE:
            process.stderr.close()
        if args['stdin'] is PIPE:
            process.stdin.close()

        process.wait()


def run_shell_command(command, stdin=None, **kwargs):
    """
    Runs a single command in shell and returns the read stdout and stderr data.

    This function waits for the process (created using ``subprocess.Popen()``)
    to exit. Effectively it wraps ``run_interactive_shell_command()`` and uses
    ``communicate()`` on the process.

    See also ``run_interactive_shell_command()``.

    :param command: The command to run on shell. This parameter can either
                    be a sequence of arguments that are directly passed to
                    the process or a string. A string gets splitted beforehand
                    using ``shlex.split()``.
    :param stdin:   Initial input to send to the process.
    :param kwargs:  Additional keyword arguments to pass to
                    ``subprocess.Popen`` that is used to spawn the process.
    :return:        A tuple with ``(stdoutstring, stderrstring)``.
    """
    with run_interactive_shell_command(command, **kwargs) as p:
        ret = p.communicate(stdin)
    return ret


def get_shell_type():  # pragma: no cover
    """
    Finds the current shell type based on the outputs of common pre-defined
    variables in them. This is useful to identify which sort of escaping
    is required for strings.

    :return: The shell type. This can be either "powershell" if Windows
             Powershell is detected, "cmd" if command prompt is been
             detected or "sh" if it's neither of these.
    """
    out = run_shell_command('echo $host.name', shell=True)[0]
    if out.strip() == 'ConsoleHost':
        return 'powershell'
    out = run_shell_command('echo $0', shell=True)[0]
    if out.strip() == '$0':
        return 'cmd'
    return 'sh'
